@@ -7,8 +7,8 @@
  * development), English, is working for the PluralForm javascript module.
  */
 
-const { PluralForm } = ChromeUtils.import(
-  "resource://gre/modules/PluralForm.jsm"
+const { PluralForm } = ChromeUtils.importESModule(
+  "resource://gre/modules/PluralForm.sys.mjs"
 );
 
 function run_test() {
